package cs3500.connectn.consoleui;

import cs3500.connectn.model.Player;

/**
 * Model for backing the {@link View}. This abstract class is designed to remove
 * the direct dependency of the view on the model; instead, the controller
 * implements this interface with information from the view and then passes the
 * {@code ViewModel} object for the view to render concretely.
 *
 * <p> See {@link ViewModel#fromArray(Player[][])} for a static factory for
 * constructing a quick-and-dirty {@code ViewModel} from a backing array.
 *
 * <p> <strong>Design Note.</strong> Is it worth having this?
 */
public abstract class ViewModel {
  /**
   * Constructs a new {@code ViewModel} with the given dimensions.
   *
   * @param width the width of the grid
   * @param height the height of the grid
   */
  protected ViewModel(int width, int height) {
    this.width = width;
    this.height = height;
  }

  /**
   * The width of the grid.
   *
   * @return the width
   */
  public final int getWidth() { return width; }
  private final int width;

  /**
   * The height of the grid.
   *
   * @return the height
   */
  public final int getHeight() { return height; }
  private final int height;

  /**
   * Gets the player whose token is at the given column and row. The coordinates
   * are zero-based and start in the lower left. Returns {@code null} if there
   * is no token in the given position.
   *
   * @param x the column coordinate ({@code 0 <= x < width})
   * @param y the row coordinate ({@code 0 <= y < height})
   * @return the player in the given position, or {@code null}
   * @throws IndexOutOfBoundsException if (x, y) is out of bounds
   */
  public abstract Player getPlayerAt(int x, int y);

  /**
   * Creates a new {@code ViewModel} backed by the given column-major array.
   * Changes to {@code grid} will appear in the returned {@code ViewModel} as
   * well.
   *
   * <p>
   * <strong>UNCHECKED PRECONDITIONS:</strong>
   * <ul>
   *   <li>{@code grid.length > 0}
   *   <li>{@code grid[0].length > 0}
   *   <li>for all in-bounds rows <em>y</em>,
   *          {@code grid[}<em>y</em>{@code ].length == grid[0].length}
   * </ul>
   *
   * @param grid the grid of player tokens, column-major, top-to-bottom
   * @return the new {@code ViewModel}
   */
  public static ViewModel fromArray(Player[][] grid) {
    return new ViewModel(grid[0].length, grid.length) {
      @Override public Player getPlayerAt(int x, int y) {
        return grid[grid.length - y][x];
      }
    };
  }
}
